#!/usr/bin/env python
#
# Semaphore: A simple (rule-based) bot library for Signal Private Messenger.
# Copyright (C) 2020 Lazlo Westerhof <semaphore@lazlo.me>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
"""This module contains an object that represents a bot job."""

from datetime import datetime
from typing import Optional

from dateutil.relativedelta import relativedelta

from .message import Message
from .reply import Reply


class Job(object):
    """This object represents a bot job."""

    def __init__(self, handler, context, repeat=False, monthly=False, interval=None):
        """Initialize job."""
        self._handler = handler
        self._context = context
        self._repeat: bool = repeat
        self._interval: int = interval
        self._monthly: bool = monthly
        self._remove: bool = False

    def get_message(self) -> Message:
        """Get the message of this job."""
        return self._context.message

    def get_interval(self) -> float:
        """Get the interval of the (repeating) job."""
        if self._repeat:
            if self._monthly:
                now = datetime.now()
                next_month = now + relativedelta(months=+1)
                interval = next_month.timestamp() - now.timestamp()
                return interval
            else:
                return self._interval
        else:
            return 0.0

    def is_repeating(self) -> bool:
        """Check if the job is repeating."""
        return self._repeat

    def schedule_removal(self) -> None:
        """Schedule the job for removal from the job queue."""
        self._remove = True

    def remove(self) -> bool:
        """Check if job should be removed."""
        return self._remove

    async def run(self) -> Optional[Reply]:
        """Run the job by calling the handler."""
        return await self._handler(self._context)



#!/usr/bin/env python
#
# Semaphore: A simple (rule-based) bot library for Signal Private Messenger.
# Copyright (C) 2020 Lazlo Westerhof <semaphore@lazlo.me>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
"""This module contains an object that represents a Signal message queue."""
import json
import logging
from typing import AsyncIterable, Optional

from .attachment import Attachment
from .data_message import DataMessage
from .group import Group
from .message import Message
from .message_sender import MessageSender
from .sticker import Sticker
from .sticker_pack import StickerPack
from .socket import Socket


class MessageReceiver:
    """This object represents a Signal message queue."""

    def __init__(self, socket: Socket, sender: MessageSender):
        """Initialize message receiver."""
        self._socket: Socket = socket
        self._sender: MessageSender = sender
        self.log = logging.getLogger(__name__)

    async def receive(self) -> AsyncIterable[Message]:
        """Receive messages and return as Iterator."""
        async for raw_line in self._socket.read():
            line: str = raw_line.decode()
            self.log.debug(f'Socket receive: {line}')

            # Load Signal message wrapper
            try:
                message_wrapper = json.loads(line)
            except json.JSONDecodeError:
                continue

            # Only handle messages.
            if message_wrapper["type"] != "message":
                continue

            try:
                message = message_wrapper["data"]
                data_message: Optional[DataMessage] = None
                data = message.get("dataMessage")
                if data:
                    group: Optional[Group] = None
                    if data.get("group"):
                        group = Group(
                            group_id=data["group"].get("groupId"),
                            name=data["group"].get("name"),
                            group_type=data["group"].get("type"),
                        )

                    sticker_data = data.get("sticker")
                    sticker: Optional[Sticker] = None
                    if sticker_data:
                        pack = StickerPack(
                            pack_id=sticker_data["packID"],
                            pack_key=sticker_data["packKey"],
                        )
                        sticker = Sticker(
                            sticker_id=sticker_data["stickerID"],
                            pack=pack,
                        )

                    data_message = DataMessage(
                        timestamp=data["timestamp"],
                        body=data.get("body", ""),
                        expires_in_seconds=data["expiresInSeconds"],
                        attachments=[
                            Attachment(
                                content_type=attachment["contentType"],
                                id=attachment["id"],
                                size=attachment["size"],
                                stored_filename=attachment["storedFilename"],
                                width=attachment["width"],
                                height=attachment["height"],
                            )
                            for attachment in data.get("attachments", [])
                        ],
                        group=group,
                        sticker=sticker,
                    )

                yield Message(
                    username=message["username"],
                    source=message["source"].get("number"),
                    envelope_type=message["type"],
                    timestamp=message["timestamp"],
                    timestamp_iso=message["timestampISO"],
                    server_timestamp=message["serverTimestamp"],
                    source_device=message.get("sourceDevice"),
                    uuid=message.get("uuid"),
                    relay=message.get("relay"),
                    has_legacy_message=message.get("hasLegacyMessage"),
                    is_receipt=message.get("isReceipt"),
                    is_unidentified_sender=message.get("isUnidentifiedSender"),
                    data_message=data_message,
                    sender=self._sender,
                )
            except Exception as exc:
                self.log.error(
                    f"Could not receive message: {json.dumps(message)}",
                    exc_info=exc,
                )



#!/usr/bin/env python
#
# Semaphore: A simple (rule-based) bot library for Signal Private Messenger.
# Copyright (C) 2020 Lazlo Westerhof <semaphore@lazlo.me>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
"""This module contains an object that represents the context of a chat."""
from typing import Any, Dict, Match, TYPE_CHECKING

import attr

from .job_queue import JobQueue
from .message import Message


# Resolve circular import.
if TYPE_CHECKING:
    from .bot import Bot


@attr.s(auto_attribs=True)
class ChatContext():
    """This object represents the context of a chat."""

    message: Message
    match: Match
    job_queue: JobQueue
    bot: 'Bot'
    data: Dict[str, Any] = {}



#!/usr/bin/env python
#
# Semaphore: A simple (rule-based) bot library for Signal Private Messenger.
# Copyright (C) 2020 Lazlo Westerhof <semaphore@lazlo.me>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
"""This module contains an object that represents a Signal message."""
from typing import Optional, TYPE_CHECKING

import attr

from .data_message import DataMessage
from .reply import Reply
if TYPE_CHECKING:
    # resolve circular imports
    from .message_sender import MessageSender


@attr.s(auto_attribs=True, frozen=True)
class Message:
    """This object represents a Signal message."""

    username: str
    source: str
    envelope_type: int
    timestamp: int
    timestamp_iso: str
    server_timestamp: int
    _sender: 'MessageSender'
    source_device: int = attr.ib(default=0)
    uuid: str = attr.ib(default=None)
    relay: str = attr.ib(default=None)
    has_legacy_message: bool = attr.ib(default=False)
    has_content: bool = attr.ib(default=False)
    is_receipt: bool = attr.ib(default=False)
    data_message: Optional[DataMessage] = attr.ib(default=None)
    is_unidentified_sender: bool = attr.ib(default=False)

    def empty(self) -> bool:
        """Check if the message is not empty."""
        if self.data_message:
            if self.data_message.body != "":
                return False
        return True

    def get_redacted_source(self) -> str:
        """Return the message source redacted."""
        return f"+********{self.source[-3:]}"

    def get_body(self):
        """Check if the message is not empty."""
        if self.empty():
            return ""
        else:
            return self.data_message.body

    def get_sticker(self):
        """Return the message's sticker if there is one."""
        return self.data_message.sticker

    def get_group_id(self) -> Optional[str]:
        """Get group id if message is a group message."""
        if self.data_message and self.data_message.group:
            return self.data_message.group.group_id
        else:
            return None

    async def reply(self, *args, **kwargs) -> None:
        """Send a reply to the message.

        The args are the same as for the Reply constructor.
        """
        await self._sender.send_message(self, Reply(*args, **kwargs))

    async def mark_delivered(self) -> None:
        """Mark the message as delivered."""
        await self._sender.mark_delivered(self)

    async def mark_read(self) -> None:
        """Mark the message as read."""
        await self._sender.mark_read(self)



#!/usr/bin/env python
#
# Semaphore: A simple (rule-based) bot library for Signal Private Messenger.
# Copyright (C) 2020 Lazlo Westerhof <semaphore@lazlo.me>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
"""Semaphore: A simple (rule-based) bot library for Signal Private Messenger."""
from .meta import *

from .attachment import Attachment
from .bot import Bot, StopPropagation
from .chat_context import ChatContext
from .data_message import DataMessage
from .group import Group
from .job import Job
from .job_queue import JobQueue
from .message import Message
from .message_receiver import MessageReceiver
from .message_sender import MessageSender
from .reply import Reply
from .socket import Socket



#!/usr/bin/env python3
# SPDX-License-Identifier: CC0-1.0

"""PriorityQueue implementation for anyio."""
from heapq import heappush, heappop

import anyio
from anyio import WouldBlock


class PriorityQueue:
    """PriorityQueue implementation for anyio."""

    def __init__(self):
        """Initialize PriorityQueue."""
        self._not_empty = anyio.create_event()
        self._queue = []
        self._counter = 0

    async def put_nowait(self, prio, item):
        """Put an item into the queue without blocking."""
        heappush(self._queue, (prio, self._counter, item))
        self._counter += 1
        if not self._not_empty.is_set():
            await self._not_empty.set()

    def get_nowait(self):
        """Remove and return an item if one is immediately available.

        If not, raise WouldBlock.
        """
        if not self._not_empty.is_set():
            raise WouldBlock

        prio, _, item = heappop(self._queue)

        if not self._queue:
            self._not_empty = anyio.create_event()

        return prio, item

    async def get(self):
        """Remove and return an item from the queue.

        If the queue is empty, wait until an item is available.
        """
        while True:
            await self._not_empty.wait()
            try:
                return self.get_nowait()
            except WouldBlock:
                pass



#!/usr/bin/env python
#
# Semaphore: A simple (rule-based) bot library for Signal Private Messenger.
# Copyright (C) 2020 Lazlo Westerhof <semaphore@lazlo.me>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
"""This module contains an object that represents a Signal data message."""
from typing import List, Optional, TYPE_CHECKING

import attr

if TYPE_CHECKING:
    from .sticker import Sticker


@attr.s(auto_attribs=True, frozen=True)
class StickerPack:
    """This object represents a Signal sticker pack."""

    pack_id: str
    pack_key: str
    stickers: List['Sticker'] = attr.ib(factory=list)
    # inbound partial sticker packs do not have title and author set
    title: Optional[str] = None
    author: Optional[str] = None

    @property
    def url(self) -> str:
        """Return the pack's URL."""
        return f'https://signal.art/addstickers/#pack_id={self.pack_id}&pack_key={self.pack_key}'  # noqa: E501



#!/usr/bin/env python
#
# Semaphore: A simple (rule-based) bot library for Signal Private Messenger.
# Copyright (C) 2020 Lazlo Westerhof <semaphore@lazlo.me>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
"""This module contains a class that handles sending bot messages."""
from typing import Any, Dict

from .message import Message
from .reply import Reply
from .socket import Socket


class MessageSender:
    """This class handles sending bot messages."""

    def __init__(self, username: str, socket: Socket):
        """Initialize message sender."""
        self._username: str = username
        self._socket: Socket = socket

    async def _send(self, message: Dict) -> None:
        await self._socket.send(message)

    async def send_message(self, message: Message, reply: Reply) -> None:
        """
        Send the bot message.

        message: The original message replying to.
        reply:   The reply to send.
        """
        # Mark message as read before replying.
        if reply.mark_read:
            await message.mark_read()

        # Construct reply message.
        bot_message: Dict[str, Any] = {"type": "react",
                                       "username": self._username}

        # Add message recipient.
        if message.get_group_id():
            bot_message["recipientGroupId"] = message.get_group_id()
        else:
            bot_message["recipientAddress"] = {"number": message.source}

        if reply.reaction:
            bot_message["type"] = "react"
            bot_message["reaction"] = {"emoji": reply.body,
                                       "targetAuthor": {"number": message.source},
                                       "targetSentTimestamp": message.timestamp}
        else:
            bot_message["type"] = "send"
            bot_message["messageBody"] = reply.body

        # Add attachments to message.
        if reply.attachments:
            bot_message["attachments"] = reply.attachments

        # Add quote to message.
        if reply.quote:
            quote = {"id": message.timestamp,
                     "author": message.source,
                     "text": message.get_body()}
            bot_message["quote"] = quote

        await self._send(bot_message)

    async def mark_delivered(self, message: Message) -> None:
        """
        Mark a Signal message you received as delivered.

        message: The Signal message you received.
        """
        await self._send({
            "type": "mark_delivered",
            "username": self._username,
            "recipientAddress": {"number": message.source},
            "timestamps": [message.timestamp],
        })

    async def mark_read(self, message: Message) -> None:
        """
        Mark a Signal message you received as read.

        message: The Signal message you received.
        """
        await self._send({
            "type": "mark_read",
            "username": self._username,
            "recipientAddress": {"number": message.source},
            "timestamps": [message.timestamp],
        })



#!/usr/bin/env python
#
# Semaphore: A simple (rule-based) bot library for Signal Private Messenger.
# Copyright (C) 2020 Lazlo Westerhof <semaphore@lazlo.me>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
"""This module contains an object that represents a bot job queue."""
import logging
from time import time
from typing import Callable

from anyio import sleep, WouldBlock

from .queue import PriorityQueue
from .job import Job


class JobQueue:
    """"This object represents a bot job queue."""

    def __init__(self, sender):
        """Initialize job queue."""
        self._queue = PriorityQueue()
        self._sender = sender

        self.log = logging.getLogger(__name__)

    async def run_once(
        self,
        timestamp: float,
        callback: Callable,
        context,
    ) -> Job:
        """Add a job to the queue that runs once."""
        job = Job(callback, context)
        await self._queue.put_nowait(timestamp, job)
        self.log.info(f"Put job ({id(job)}) in the queue")
        return job

    async def run_repeating(self,
                            timestamp: float,
                            callback: Callable,
                            context,
                            interval: int) -> Job:
        """Add a job to the queue that runs repeating."""
        job = Job(callback, context, repeat=True, interval=interval)
        await self._queue.put_nowait(timestamp, job)
        self.log.info(f"Put repeating job ({id(job)}) in the queue")
        return job

    async def run_daily(self,
                        timestamp: float,
                        callback: Callable,
                        context) -> Job:
        """Add a job to the queue that runs daily."""
        interval = 60 * 60 * 24  # Day
        job = Job(callback, context, repeat=True, interval=interval)
        await self._queue.put_nowait(timestamp, job)
        self.log.info(f"Put daily job ({id(job)}) in the queue")
        return job

    async def run_monthly(self,
                          timestamp: float,
                          callback: Callable,
                          context) -> Job:
        """Add a job to the queue that runs monthly."""
        job = Job(callback, context, repeat=True, monthly=True)
        await self._queue.put_nowait(timestamp, job)
        self.log.info(f"Put monthly job ({id(job)}) in the queue")
        return job

    async def start(self) -> None:
        """Run all the jobs in the queue that are due."""
        self.log.info("Job queue started")

        while True:
            now = time()

            try:
                timestamp, job = self._queue.get_nowait()
            except WouldBlock:
                await sleep(1)
                continue

            if timestamp > now:
                await self._queue.put_nowait(timestamp, job)
                await sleep(1)
                continue

            if job.remove():
                self.log.info(f"Removed job ({id(job)}) from queue")
                continue

            self.log.info(f"Running job ({id(job)})")
            message = job.get_message()
            try:
                reply = await job.run()
                if reply:
                    await self._sender.send_message(message, reply)
                    self.log.info(f"Reply for job ({id(job)}) sent "
                                  f"to {message.get_redacted_source()}")
            except Exception as exc:
                self.log.warning(f"Sending reply for message ({id(message)}) "
                                 f"to {message.get_redacted_source()} failed",
                                 exc_info=exc)
                continue

            if job.is_repeating():
                interval = job.get_interval()
                await self._queue.put(now + interval, job)
                self.log.info(f"Added repeating job ({id(job)}) to the queue")



#!/usr/bin/env python
#
# Semaphore: A simple (rule-based) bot library for Signal Private Messenger.
# Copyright (C) 2020 Lazlo Westerhof <semaphore@lazlo.me>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
"""This module contains an object that represents a Signal data message."""
from typing import List, Optional

import attr

from .attachment import Attachment
from .sticker import Sticker
from .group import Group


@attr.s(auto_attribs=True, frozen=True)
class DataMessage:
    """This object represents a Signal data message."""

    timestamp: int
    body: str
    expires_in_seconds: int = attr.ib(default=0)
    attachments: List[Attachment] = attr.ib(factory=list)
    group: Optional[Group] = attr.ib(default=None)
    sticker: Optional[Sticker] = attr.ib(default=None)



#!/usr/bin/env python
#
# Semaphore: A simple (rule-based) bot library for Signal Private Messenger.
# Copyright (C) 2020 Lazlo Westerhof <semaphore@lazlo.me>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
"""This module contains an object that represents a signald socket."""
import json
import logging
from typing import AsyncIterable, List

import anyio
import anyio.abc


class Socket:
    """This object represents a signald socket."""

    def __init__(self, username: str, socket_path: str = "/var/run/signald/signald.sock"):
        """Initialize socket."""
        self._username: str = username
        self._socket_path: str = socket_path
        self._socket: anyio.abc.SocketStream

        self.log = logging.getLogger(__name__)

    async def __aenter__(self) -> 'Socket':
        """Connect to the socket."""
        self._socket = await (await anyio.connect_unix(self._socket_path)).__aenter__()
        self.log.info(f"Connected to socket ({self._socket_path})")
        await self.send({"type": "subscribe", "username": self._username})
        self.log.info(f"Bot attempted to subscribe to +********{self._username[-3:]}")
        return self

    async def __aexit__(self, *excinfo):
        """Disconnect from the internal socket."""
        return await self._socket.__aexit__(*excinfo)

    async def read(self) -> AsyncIterable[bytes]:
        """Read a socket, line by line."""
        buffer: List[bytes] = []
        while True:
            try:
                char = await self._socket.receive(1)
            except anyio.EndOfStream:
                raise ConnectionResetError("Connection was reset")
            if char == b"\n":
                yield b"".join(buffer)
                buffer = []
            else:
                buffer.append(char)

    async def send(self, message: dict) -> None:
        """Send message to socket."""
        serialized = json.dumps(message)
        self.log.debug(f"Socket send: {serialized}")
        await self._socket.send(serialized.encode("utf8") + b"\n")



#!/usr/bin/env python
#
# Semaphore: A simple (rule-based) bot library for Signal Private Messenger.
# Copyright (C) 2020 Lazlo Westerhof <semaphore@lazlo.me>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
"""This module contains an object that represents a Signal Private Messenger bot."""
import logging
import re
import threading
from datetime import datetime
from typing import Any, Callable, Dict, List, Pattern, TYPE_CHECKING

import anyio

from .chat_context import ChatContext
from .job_queue import JobQueue
from .message import Message
from .message_receiver import MessageReceiver
from .message_sender import MessageSender
from .socket import Socket


class StopPropagation(Exception):
    """Raise this to prevent further handlers from running on this message."""


class Bot:
    """This object represents a simple (rule-based) Signal Private Messenger bot."""

    def __init__(self,
                 username: str,
                 logging_level=logging.INFO,
                 socket_path="/var/run/signald/signald.sock"):
        """Initialize bot."""
        self._username: str = username
        self._socket_path: str = socket_path
        self._receiver: MessageReceiver
        self._sender: MessageSender
        self._job_queue: JobQueue
        self._handlers: List = []
        self._chat_context: Dict[str, ChatContext] = {}

        threading.current_thread().name = 'bot'
        logging.basicConfig(
            format='%(asctime)s %(threadName)s: [%(levelname)s] %(message)s',
            level=logging_level
        )
        self.log = logging.getLogger(__name__)

        self._socket: Socket

    def register_handler(self, regex: Pattern, func: Callable) -> None:
        """Register a chat handler with a regex."""
        if not isinstance(regex, type(re.compile(""))):
            regex = re.compile(regex, re.UNICODE)

        self._handlers.append((regex, func))
        self.log.info(f"Handler <{func.__name__}> registered ('{regex.pattern}')")

    def handler(self, regex: Pattern):
        """Decorator to register handlers."""
        def decorator(func: Callable):
            self.register_handler(regex, func)
            return func
        return decorator

    async def _handle_message(self, message: Message, func: Callable, match) -> None:
        """Handle a matched message."""
        message_id = id(message)
        message_source = message.get_redacted_source()

        # Retrieve or create chat context.
        if self._chat_context.get(message.source, False):
            context = self._chat_context[message.source]
            context.message = message
            context.match = match
            self.log.debug(f"Chat context exists for {message_source}")
        else:
            context = ChatContext(message, match, self._job_queue, self)
            self.log.debug(f"Chat context created for {message_source}")

        # Process received message and send reply.
        try:
            await func(context)
            self._chat_context[message.source] = context
            self.log.debug(f"Message ({message_id}) processed by handler {func.__name__}")
        except StopPropagation:
            raise
        except Exception as exc:
            self.log.error(
                f"Processing message ({message_id}) by {func.__name__} failed",
                exc_info=exc,
            )

    async def _match_message(self, message: Message) -> None:
        """Match an incoming message against a handler."""
        message_id = id(message)
        message_source = message.get_redacted_source()

        # Mark message as delivered.
        await message.mark_delivered()
        self.log.debug(f"Message ({message_id}) received from {message_source}")
        self.log.debug(str(message))

        # Loop over all registered handlers.
        for regex, func in self._handlers:
            # Match message text against handlers.
            match = re.search(regex, message.get_body())
            if not match:
                self.log.debug(f'skipping {func.__name__}')
                continue

            self.log.debug(
                f"Message matched against handler <{func.__name__}> ('{regex.pattern}')"
            )

            try:
                await self._handle_message(message, func, match)
            except StopPropagation:
                break

    async def __aenter__(self) -> 'Bot':
        """Connect to the bot's internal socket."""
        self._socket = await Socket(self._username, self._socket_path).__aenter__()
        return self

    async def __aexit__(self, *excinfo):
        """Disconnect from the bot's internal socket."""
        return await self._socket.__aexit__(*excinfo)

    async def start(self) -> None:
        """Start the bot event loop."""
        self.log.info("Bot started")

        self._sender = MessageSender(self._username, self._socket)
        self._receiver = MessageReceiver(self._socket, self._sender)

        async with anyio.create_task_group() as tg:
            self._job_queue = JobQueue(self._sender)
            await tg.spawn(self._job_queue.start)

            # handle incoming messages in parallel
            async for message in self._receiver.receive():
                if message.data_message is not None:
                    await tg.spawn(self._match_message, message)



#!/usr/bin/env python
import importlib
import importlib.util
from pathlib import Path
from setuptools import find_packages, setup


def import_by_path(module_name, path):
    """Import the module name and not its parent package.
    Return the module.
    """
    spec = importlib.util.spec_from_file_location(module_name, path)
    module = importlib.util.module_from_spec(spec)
    spec.loader.exec_module(module)
    return module


meta = import_by_path('meta', Path(__file__).parent / 'semaphore' / 'meta.py')

# Get the long description from the README file
with open("README.md", "r") as fh:
    long_description = fh.read()

requirements = [
    'anyio<3.0.0',
    'attrs==18.2.0',
    'attr==0.3.1',
    'python_dateutil==2.8.1',
]

setup(
    name="semaphore-bot",
    version=meta.__version__,
    author=meta.__author__,
    author_email=meta.__email__,
    description=(
        "Semaphore: A simple (rule-based) bot library for Signal Private Messenger."
    ),
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/lwesterhof/semaphore",
    license=meta.__license__,
    packages=find_packages(),
    install_requires=requirements,
    extras_require={'dev': [
        'flake8',
        'pep257',
        'mypy',
        'sphinx',
        'sphinx_rtd_theme',
    ]},
    classifiers=[
        'Development Status :: 4 - Beta',
        'Intended Audience :: Developers',
        (
            'License :: OSI Approved :: '
            'GNU Affero General Public License v3 or later (AGPLv3+)'
        ),
        'Operating System :: OS Independent',
        'Topic :: Software Development :: Libraries :: Python Modules',
        'Topic :: Communications :: Chat',
        'Topic :: Internet',
        'Programming Language :: Python',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
        'Programming Language :: Python :: 3.9',
    ],
    python_requires='>=3.6',
    project_urls={
        'Bug Reports': 'https://github.com/lwesterhof/semaphore/issues',
        'Source': 'https://github.com/lwesterhof/semaphore',
    },
)
